@@ -22,14 +22,16 @@ from tempfile import TemporaryDirectory
 from threading import Event
 from urllib.request import Request, urlopen, urljoin, quote
 from urllib.error import URLError
+import warnings
 
 from jupyter_core.paths import jupyter_config_path
 from jupyterlab_server.process import which, Process, WatchHelper, list2cmdline
 from notebook.nbextensions import GREEN_ENABLED, GREEN_OK, RED_DISABLED, RED_X
+from traitlets import HasTraits, Bool, Unicode, Instance, default
 
 from .semver import Range, gte, lt, lte, gt, make_semver
 from .jlpmapp import YARN_PATH, HERE
-from .coreconfig import _get_default_core_data
+from .coreconfig import _get_default_core_data, CoreConfig
 
 
 # The regex for expecting the webpack output.
@@ -42,6 +44,11 @@ DEV_DIR = osp.abspath(os.path.join(HERE, '..', 'dev_mode'))
 # If we are pinning the package, rename it `pin@<alias>`
 PIN_PREFIX = 'pin@'
 
+
+# Default Yarn registry used in default yarn.lock
+YARN_DEFAULT_REGISTRY = 'https://registry.yarnpkg.com'
+
+
 class ProgressProcess(Process):
 
     def __init__(self, cmd, logger=None, cwd=None, kill_event=None,
@@ -67,7 +74,7 @@ class ProgressProcess(Process):
         if kill_event and kill_event.is_set():
             raise ValueError('Process aborted')
 
-        self.logger = logger = logger or logging.getLogger('jupyterlab')
+        self.logger = _ensure_logger(logger)
         self._last_line = ''
         self.cmd = cmd
         self.logger.debug('> ' + list2cmdline(cmd))
@@ -88,7 +95,7 @@ class ProgressProcess(Process):
         proc = self.proc
         kill_event = self._kill_event
         import itertools
-        spinner = itertools.cycle(['-', '/', '|', '\\'])
+        spinner = itertools.cycle(['-', '\\', '|', '/'])
         while proc.poll() is None:
             sys.stdout.write(next(spinner))   # write the next character
             sys.stdout.flush()                # flush stdout buffer (actual character display)
@@ -287,7 +294,75 @@ def watch_dev(logger=None):
     return package_procs + [wp_proc]
 
 
-def watch(app_dir=None, logger=None, core_config=None):
+class AppOptions(HasTraits):
+    """Options object for build system"""
+
+    def __init__(self, logger=None, core_config=None, **kwargs):
+        if core_config is not None:
+            kwargs['core_config'] = core_config
+        if logger is not None:
+            kwargs['logger'] = logger
+
+        # use the default if app_dir is empty
+        if 'app_dir' in kwargs and not kwargs['app_dir']:
+            kwargs.pop('app_dir')
+
+        super(AppOptions, self).__init__(**kwargs)
+
+    app_dir = Unicode(help='The application directory')
+
+    use_sys_dir = Bool(
+        True,
+        help=('Whether to shadow the default app_dir if that is set to a '
+             'non-default value'))
+
+    logger = Instance(logging.Logger, help='The logger to use')
+
+    core_config = Instance(CoreConfig, help='Configuration for core data')
+
+    kill_event = Instance(Event, args=(), help='Event for aborting call')
+
+    registry = Unicode(help="NPM packages registry URL")
+
+    @default('logger')
+    def _default_logger(self):
+        return logging.getLogger('jupyterlab')
+
+    # These defaults need to be dynamic to pick up
+    # any changes to env vars:
+    @default('app_dir')
+    def _default_app_dir(self):
+        return get_app_dir()
+
+    @default('core_config')
+    def _default_core_config(self):
+        return CoreConfig()
+
+    @default('registry')
+    def _default_registry(self):
+        config = _yarn_config(self.logger)["yarn config"]
+        return config.get("registry", YARN_DEFAULT_REGISTRY)
+
+
+def _ensure_options(options, **kwargs):
+    """Helper to use deprecated kwargs for AppOption"""
+    # Filter out default-value kwargs
+    kwargs = dict(filter(lambda item: item[1] is not None, kwargs.items()))
+    # Warn for deprecated kwargs usage
+    if kwargs:
+        warnings.warn(
+            "Direct keyword args to jupyterlab.commands functions are "
+            "deprecated, use the options argument instead: %r" % (kwargs,),
+            DeprecationWarning)
+    if options is None:
+        return AppOptions(**kwargs)
+    # Also support mixed use of options and kwargs:
+    opt_args = {name: getattr(options, name) for name in options.trait_names()}
+    kwargs.update(**opt_args)
+    return AppOptions(**kwargs)
+
+
+def watch(app_dir=None, logger=None, core_config=None, app_options=None):
     """Watch the application.
 
     Parameters
@@ -301,59 +376,64 @@ def watch(app_dir=None, logger=None, core_config=None):
     -------
     A list of processes to run asynchronously.
     """
-    logger = _ensure_logger(logger)
-    _node_check(logger)
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config)
+    _node_check(app_options.logger)
+    handler = _AppHandler(app_options)
     return handler.watch()
 
 
 
-def install_extension(extension, app_dir=None, logger=None, core_config=None, pin=None):
+def install_extension(extension, app_dir=None, logger=None, core_config=None, pin=None, app_options=None):
     """Install an extension package into JupyterLab.
 
     The extension is first validated.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    logger = _ensure_logger(logger)
-    _node_check(logger)
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config)
+    _node_check(app_options.logger)
+    handler = _AppHandler(app_options)
     return handler.install_extension(extension, pin=pin)
 
 
-def uninstall_extension(name=None, app_dir=None, logger=None, all_=False, core_config=None):
+def uninstall_extension(name=None, app_dir=None, logger=None, all_=False, core_config=None, app_options=None):
     """Uninstall an extension by name or path.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    logger = _ensure_logger(logger)
-    _node_check(logger)
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config)
+    _node_check(app_options.logger)
+    handler = _AppHandler(app_options)
     if all_ is True:
         return handler.uninstall_all_extensions()
     return handler.uninstall_extension(name)
 
 
-def update_extension(name=None, all_=False, app_dir=None, logger=None, core_config=None):
+def update_extension(name=None, all_=False, app_dir=None, logger=None, core_config=None, app_options=None):
     """Update an extension by name, or all extensions.
-
     Either `name` must be given as a string, or `all_` must be `True`.
     If `all_` is `True`, the value of `name` is ignored.
-
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    logger = _ensure_logger(logger)
-    _node_check(logger)
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config)
+    _node_check(app_options.logger)
+    handler = _AppHandler(app_options)
     if all_ is True:
         return handler.update_all_extensions()
     return handler.update_extension(name)
 
 
-def clean(app_dir=None, logger=None):
+def clean(app_dir=None, logger=None, app_options=None):
     """Clean the JupyterLab application directory."""
-    logger = _ensure_logger(logger)
-    app_dir = app_dir or get_app_dir()
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger)
+    handler = _AppHandler(app_options)
+    logger = app_options.logger
+    app_dir = app_options.app_dir
     logger.info('Cleaning %s...', app_dir)
     if app_dir == pjoin(HERE, 'dev'):
         raise ValueError('Cannot clean the dev app')
@@ -368,96 +448,105 @@ def clean(app_dir=None, logger=None):
 
 def build(app_dir=None, name=None, version=None, static_url=None,
           logger=None, command='build:prod', kill_event=None,
-          clean_staging=False, core_config=None):
+          clean_staging=False, core_config=None, app_options=None):
     """Build the JupyterLab application.
     """
-    logger = _ensure_logger(logger)
-    _node_check(logger)
-    handler = _AppHandler(app_dir, logger, kill_event=kill_event, core_config=core_config)
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config)
+    _node_check(app_options.logger)
+    handler = _AppHandler(app_options)
     return handler.build(name=name, version=version, static_url=static_url,
                          command=command, clean_staging=clean_staging)
 
 
-def get_app_info(app_dir=None, logger=None, core_config=None):
+def get_app_info(app_dir=None, logger=None, core_config=None, app_options=None):
     """Get a dictionary of information about the app.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.info
 
 
-def enable_extension(extension, app_dir=None, logger=None, core_config=None):
+def enable_extension(extension, app_dir=None, logger=None, core_config=None, app_options=None):
     """Enable a JupyterLab extension.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.toggle_extension(extension, False)
 
 
-def disable_extension(extension, app_dir=None, logger=None, core_config=None):
+def disable_extension(extension, app_dir=None, logger=None, core_config=None, app_options=None):
     """Disable a JupyterLab package.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.toggle_extension(extension, True)
 
 
-def check_extension(extension, app_dir=None, installed=False, logger=None, core_config=None):
+def check_extension(extension, app_dir=None, installed=False, logger=None, core_config=None, app_options=None):
     """Check if a JupyterLab extension is enabled or disabled.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.check_extension(extension, installed)
 
 
-def build_check(app_dir=None, logger=None, core_config=None):
+def build_check(app_dir=None, logger=None, core_config=None, app_options=None):
     """Determine whether JupyterLab should be built.
 
     Returns a list of messages.
     """
-    logger = _ensure_logger(logger)
-    _node_check(logger)
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    app_options = _ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config)
+    _node_check(app_options.logger)
+    handler = _AppHandler(app_options)
     return handler.build_check()
 
 
-def list_extensions(app_dir=None, logger=None, core_config=None):
+def list_extensions(app_dir=None, logger=None, core_config=None, app_options=None):
     """List the extensions.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.list_extensions()
 
 
-def link_package(path, app_dir=None, logger=None, core_config=None):
+def link_package(path, app_dir=None, logger=None, core_config=None, app_options=None):
     """Link a package against the JupyterLab build.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.link_package(path)
 
 
-def unlink_package(package, app_dir=None, logger=None, core_config=None):
+def unlink_package(package, app_dir=None, logger=None, core_config=None, app_options=None):
     """Unlink a package from JupyterLab by path or name.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
     """
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.unlink_package(package)
 
 
-def get_app_version(app_dir=None, core_config=None):
+def get_app_version(app_dir=None, core_config=None, app_options=None):
     """Get the application version."""
-    app_dir = app_dir or get_app_dir()
-    handler = _AppHandler(app_dir, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, core_config=core_config))
     return handler.info['version']
 
 
-def get_latest_compatible_package_versions(names, app_dir=None, logger=None, core_config=None):
+def get_latest_compatible_package_versions(names, app_dir=None, logger=None, core_config=None, app_options=None):
     """Get the latest compatible version of a list of packages.
     """
-    app_dir = app_dir or get_app_dir()
-    handler = _AppHandler(app_dir, logger, core_config=core_config)
+    handler = _AppHandler(_ensure_options(
+        app_options, app_dir=app_dir, logger=logger, core_config=core_config))
     return handler.latest_compatible_package_versions(names)
 
 
@@ -481,19 +570,16 @@ def read_package(target):
 
 class _AppHandler(object):
 
-    def __init__(self, app_dir, logger=None, kill_event=None, core_config=None):
+    def __init__(self, options):
         """Create a new _AppHandler object
         """
-        self.app_dir = app_dir or get_app_dir()
-        self.sys_dir = get_app_dir()
-        self.logger = _ensure_logger(logger)
-        self.core_data = (
-            core_config._data if core_config else _get_default_core_data()
-        )
+        self.app_dir = options.app_dir
+        self.sys_dir = get_app_dir() if options.use_sys_dir else self.app_dir
+        self.logger = options.logger
+        self.core_data = options.core_config._data
         self.info = self._get_app_info()
-        self.kill_event = kill_event or Event()
-        # TODO: Make this configurable
-        self.registry = 'https://registry.npmjs.org'
+        self.kill_event = options.kill_event
+        self.registry = options.registry
 
     def install_extension(self, extension, existing=None, pin=None):
         """Install an extension package into JupyterLab.
@@ -554,7 +640,7 @@ class _AppHandler(object):
                         'prod')
         command = ':'.join(parts)
 
-        self.logger.info('Building jupyterlab assets (%s)'%command)
+        self.logger.info('Building jupyterlab assets (%s)' % command)
 
         # Set up the build directory.
         app_dir = self.app_dir
@@ -1083,9 +1169,16 @@ class _AppHandler(object):
             json.dump(data, fid, indent=4)
 
         # copy known-good yarn.lock if missing
-        lock_path = pjoin(staging, 'yarn.lock')
-        if not osp.exists(lock_path):
-            shutil.copy(pjoin(HERE, 'staging', 'yarn.lock'), lock_path)
+        lock_path = pjoin(staging, 'yarn.lock')        
+        lock_template = pjoin(HERE, 'staging', 'yarn.lock')
+        if self.registry != YARN_DEFAULT_REGISTRY:  # Replace on the fly the yarn repository see #3658
+            with open(lock_template, encoding='utf-8') as f:
+                template = f.read()
+            template = template.replace(YARN_DEFAULT_REGISTRY, self.registry.strip("/"))
+            with open(lock_path, 'w', encoding='utf-8') as f:
+                f.write(template)
+        elif not osp.exists(lock_path):
+            shutil.copy(lock_template, lock_path)
 
     def _get_package_template(self, silent=False):
         """Get the template the for staging package.json file.
@@ -1109,6 +1202,7 @@ class _AppHandler(object):
         # Handle local extensions.
         for (key, source) in local.items():
             jlab['linkedPackages'][key] = source
+            data['resolutions'][key] = source
 
         # Handle linked packages.
         for (key, item) in linked.items():
@@ -1116,6 +1210,7 @@ class _AppHandler(object):
             path = pjoin(path, item['filename'])
             data['dependencies'][key] = format_path(path)
             jlab['linkedPackages'][key] = item['source']
+            data['resolutions'][key] = format_path(path)
 
         # Handle extensions
         compat_errors = self._get_extension_compat()
@@ -1144,10 +1239,11 @@ class _AppHandler(object):
         for item in self.info['uninstalled_core']:
             if item in jlab['extensions']:
                 data['jupyterlab']['extensions'].pop(item)
-            else:
+            elif item in jlab['mimeExtensions']:
                 data['jupyterlab']['mimeExtensions'].pop(item)
             # Remove from dependencies as well.
-            data['dependencies'].pop(item)
+            if item in data['dependencies']:
+                data['dependencies'].pop(item)
 
         return data
 
@@ -1171,6 +1267,9 @@ class _AppHandler(object):
         """
         # Extract the package in a temporary directory.
         existing = data['filename']
+        if not osp.exists(pjoin(dname, existing)):
+            existing = ''
+
         with TemporaryDirectory() as tempdir:
             info = self._extract_package(source, tempdir)
 
@@ -1180,7 +1279,7 @@ class _AppHandler(object):
 
             shutil.move(info['path'], pjoin(dname, info['filename']))
 
-        # Remove the existing tarball and return the new file name.
+        # Remove the previous tarball and return the new file name.
         if existing:
             os.remove(pjoin(dname, existing))
 
@@ -1662,6 +1761,42 @@ def _node_check(logger):
         msg = 'Please install nodejs %s before continuing. nodejs may be installed using conda or directly from the nodejs website.' % ver
         raise ValueError(msg)
 
+def _yarn_config(logger):
+    """Get the yarn configuration.
+    
+    Returns
+    -------
+    {"yarn config": dict, "npm config": dict} if unsuccessfull the subdictionary are empty
+    """
+    configuration = {"yarn config": {}, "npm config": {}}
+    try:
+        node = which('node')
+    except ValueError:  # Node not found == user with no need for building jupyterlab
+        logger.debug("NodeJS was not found. Yarn user configuration is ignored.")
+        return configuration
+
+    try:
+        output_binary = subprocess.check_output([node, YARN_PATH, 'config', 'list', '--json'], stderr=subprocess.PIPE, cwd=HERE)
+        output = output_binary.decode('utf-8')
+        lines = iter(output.splitlines())
+        try:
+            for line in lines:
+                info = json.loads(line)
+                if info["type"] == "info":
+                    key = info["data"]
+                    inspect = json.loads(next(lines))
+                    if inspect["type"] == "inspect":
+                        configuration[key] = inspect["data"]
+        except StopIteration:
+            pass
+        logger.debug("Yarn configuration loaded.")
+    except subprocess.CalledProcessError as e:
+        logger.error("Fail to get yarn configuration. {!s}{!s}".format(e.stderr.decode('utf-8'), e.output.decode('utf-8')))
+    except Exception as e:
+        logger.error("Fail to get yarn configuration. {!s}".format(e))
+    finally:
+        return configuration
+
 
 def _ensure_logger(logger=None):
     """Ensure that we have a logger"""
